;;; Dezyne --- Dezyne command line tools
;;;
;;; Copyright © 2018, 2019, 2020, 2021, 2022, 2023, 2024 Janneke Nieuwenhuizen <janneke@gnu.org>
;;; Copyright © 2018, 2021, 2022, 2023 Rutger van Beusekom <rutger@dezyne.org>
;;; Copyright © 2018, 2020, 2022, 2023, 2024 Paul Hoogendijk <paul@dezyne.org>
;;; Copyright © 2018, 2019, 2020 Rob Wieringa <rma.wieringa@gmail.com>
;;; Copyright © 2020 Johri van Eerd <vaneerd.johri@gmail.com>
;;;
;;; This file is part of Dezyne.
;;;
;;; Dezyne is free software: you can redistribute it and/or modify it
;;; under the terms of the GNU Affero General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Dezyne is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Affero General Public License for more details.
;;;
;;; You should have received a copy of the GNU Affero General Public
;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.
;;;
;;; Commentary:
;;;
;;; Code:

(define-module (dzn ast normalize)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-9 gnu)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-71)

  #:use-module (ice-9 curried-definitions)
  #:use-module (ice-9 match)

  #:use-module (dzn ast ast)
  #:use-module (dzn ast lookup)
  #:use-module (dzn tree util)
  #:use-module (dzn ast)
  #:use-module (dzn misc)

  #:export (%normalize:short-circuit?
            add-defer-end
            add-explicit-temporaries
            add-function-port-specializations
            add-void-function-return
            add-reply-port
            binding-into-blocking
            extract-call
            global-functions->behavior
            inline-predicates
            invariant:include-guard-expressions
            make-a=>b-expression
            not-or-guards
            normalize:event
            normalize:event+illegals
            normalize:split-complex-expressions
            normalize:state
            normalize:state+illegals
            purge-data
            remove-behavior
            remove-location
            remove-otherwise
            simplify-guard-expressions
            split-complex-expressions
            split-variable
            tag-imperative-blocks))

;; Should transformation stop here?
(define %normalize:short-circuit? (make-parameter (const #f)))


;;;
;;; Utilities and predicates.
;;;
(define-method (split-variable (o <variable>))
  (let* ((default (ast:default-value o))
         (variable (clone o #:expression default))
         (name (.name o))
         (location (.location o))
         (assign (graft o (make <assign>
                            #:variable.name name
                            #:expression (.expression o)
                            #:location location))))
    (values variable assign)))

(define temp-name
  (let ((count -1))
    (lambda (o)
      (set! count (1+ count))
      (format #f "dzn_tmp~a" count))))

(define (temporaries o)
  "The mCRL2 has no support for anonymous temporary variables, Dezyne does
as do all programming languages, so we must determine which statements
require a temporary variable."
  (define action/call?
    (conjoin (disjoin (is? <action>) (is? <call>))
             (disjoin tree:parent (cute throw 'no-parent <>))))
  (define temporary?
    (conjoin (disjoin (is? <action>)
                      (is? <binary>)
                      (is? <call>))
             (disjoin tree:parent (cute throw 'no-parent <>))
             (compose pair?
                      (cute tree:collect <> action/call?))
             (disjoin
              (cute as <> <arguments>)
              (cute tree:ancestor <> <arguments>)
              (compose (cute as <> <binary>) tree:parent)
              (compose (cute tree:ancestor <> <binary>) tree:parent))))
  (cond ((or (is-a? o <call>) (is-a? o <action>))
         (tree:collect o temporary?))
        ((or (is-a? o <assign>) (is-a? o <variable>))
         (append
          (tree:collect o temporary?)
          (map .expression
               (tree:collect
                o
                (conjoin
                 (is? <not>)
                 (compose pair? (cute tree:collect <> action/call?)))))))
        ((is-a? o <if>)
         (tree:collect (.expression o) action/call?))
        ((or (as o <expression>)
             (and (is-a? o <reply>) (.expression o))
             (and (is-a? o <return>) (.expression o)))
         =>
         (cute tree:collect <> action/call?))
        (else
         '())))

(define (add-temporary? o)
  (match (temporaries o)
    ((h t ...)
     (let ((p (tree:parent h)))
       (and (not (or (as p <and>) (tree:ancestor p <and>)
                     (as p <or>) (tree:ancestor p <or>)))
            h)))
    (_
     #f)))

(define (temporaries* o)
  "In C derived languages the evaluation order of the arguments is
implementation defined, in Dezyne we assume a left to right order. This
requires the introduction of temporaries for function calls."
  (define action/call?
    (disjoin (is? <action>) (is? <call>)))
  (define >1-argument-action/call?
    (compose (cute > <> 1)
             length
             (cute filter (compose pair?
                                   (cute tree:collect <> action/call?))
                   <>)
             ast:argument*))
  (define >1-action/call?
    (compose (cute > <> 1)
             length
             (cute tree:collect <> action/call?)))
  (let* ((arguments (tree:collect
                     o
                     (conjoin (is? <arguments>)
                              (compose (cute > <> 1) length .elements)
                              >1-action/call?)))
         (arguments (filter >1-argument-action/call? arguments))
         (arguments (append-map ast:argument* arguments))
         (expressions (tree:collect
                       o
                       (conjoin (disjoin (is? <minus>) (is? <plus>))
                                >1-action/call?)))
         (expressions (append-map
                       (lambda (x)
                         (let ((expressions (list (.left x) (.right x))))
                           (filter
                            (compose pair?
                                     (cute tree:collect <> action/call?))
                            expressions)))
                       expressions)))
    (append arguments expressions)))

(define (add-temporary*? o)
  (let ((temporaries (temporaries* o)))
    (match temporaries
      ((h t ...)
       (find (negate (disjoin (is? <reference>) (is? <data>))) temporaries))
      (_
       #f))))


;;;
;;; Expressions.
;;;
(define-method (group-expression (o <binary>))
  (make <group> #:expression o))


(define-method (group-expression (o <field-test>))
  (make <group> #:expression o))

(define-method (group-expression (o <expression>))
  o)

(define (and-expressions expressions)
  (reduce (cute make <and> #:left <> #:right <>)
          (make <literal> #:value "true")
          (map group-expression expressions)))

(define (or-expressions expressions)
  (reduce (cute make <or> #:left <> #:right <>)
          (make <literal> #:value "false")
          (map group-expression expressions)))

(define (not-or-guards guards)
  (let* ((expressions (map .expression guards))
         (others (remove (is? <otherwise>) expressions))
         (expression (or-expressions others)))
    (match expression
      ((and ($ <not>) (= .expression expression)) expression)
      (_ (make <not> #:expression expression)))))

(define (and-not-guards guards)
  (let* ((expression
          (match guards
            (()
             (make <literal> #:value "true"))
            ((guard)
             (make <not> #:expression (.expression guard)))
            ((h t ...)
             (let ((expressions (map (compose (cute make <not> #:expression <>)
                                              .expression)
                                     guards)))
               (and-expressions expressions))))))
    (make <guard> #:expression expression)))

(define-method (simplify-expression (o <bool-expr>))
  (define (static? o)
    (null? (tree:collect o (disjoin (is? <action>) (is? <call>)))))
  (match o
    (($ <not>)
     (let* ((expression (.expression o))
            (e (simplify-expression expression)))
       (cond ((is-a? expression <not>)
              (simplify-expression (.expression expression)))
             ((and (is-a? expression <group>)
                   (as (simplify-expression (.expression expression))
                       <not>))
              =>
              .expression)
             ((ast:literal-true? e) (clone e #:value "false"))
             ((ast:literal-false? e) (clone e #:value "true"))
             (else (clone o #:expression e)))))
    (($ <and>)
     (let ((left (simplify-expression (.left o)))
           (right (simplify-expression (.right o))))
       (cond ((and (is-a? left <not>)
                   (static? left)
                   (ast:equal? (.expression left) right))
              (make <literal> #:value "false"))
             ((and (is-a? right <not>)
                   (static? left)
                   (ast:equal? (.expression right) left))
              (make <literal> #:value "false"))
             ((and (static? left)
                   (ast:equal? left right))
              left)
             ((ast:literal-true? left) right)
             ((ast:literal-false? left) left)
             ((ast:literal-true? right) left)
             ((and (static? left)
                   (ast:literal-false? right)) right)
             (else (clone o #:left left #:right right)))))
    (($ <or>)
     (let ((left (simplify-expression (.left o)))
           (right (simplify-expression (.right o))))
       (cond ((and (is-a? left <not>)
                   (static? left)
                   (ast:equal? (.expression left) right))
              (make <literal> #:value "true"))
             ((and (is-a? right <not>)
                   (static? left)
                   (ast:equal? (.expression right) left))
              (make <literal> #:value "true"))
             ((and (static? left)
                   (ast:equal? left right))
              left)
             ((ast:literal-true? left) left)
             ((ast:literal-false? left) right)
             ((and (static? left)
                   (ast:literal-true? right)) right)
             ((ast:literal-false? right) left)
             (else (clone o #:left left #:right right)))))
    ((? (is? <binary>))
     (clone o
            #:left (simplify-expression (.left o))
            #:right (simplify-expression (.right o))))
    (_
     o)))

(define-method (simplify-expression (o <group>))
  (let* ((expression (.expression o))
         (expression (simplify-expression expression)))
    (if (or (is-a? expression <unary>)
            (is-a? expression <field-test>)) expression
            (clone o #:expression expression))))

(define-method (simplify-expression (o <expression>))
  o)

(define-method (simplify-toplevel-expression (o <expression>))
  (let ((expression (simplify-expression o)))
    (if (is-a? expression <group>) (.expression expression)
        expression)))

(define-method (simplify-guard (o <guard>))
  (clone o #:expression (simplify-toplevel-expression (.expression o))))

(define-method (simplify-guard (o <canonical-on>))
  (clone o #:guard (simplify-guard (.guard o))))


;;;
;;; Canonical-on.
;;;
(define* (statement->canonical-on o)
  (define (imperative->canonical-ons imperative)
    (let* ((path (tree:path imperative (compose (is? <behavior>) tree:parent)))
           (blocking (find (is? <blocking>) path))
           (guards (filter (is? <guard>) path))
           (expression (and-expressions (map .expression guards)))
           (guard (make <guard> #:expression expression))
           (on (find (is? <on>) path))
           (triggers (if on (ast:trigger* on) '()))
           (behavior (tree:ancestor o <behavior>)))
      (define (make-on trigger)
        (graft behavior (make <canonical-on>
                          #:blocking (and (ast:provides? trigger) blocking)
                          #:guard guard
                          #:trigger trigger
                          #:statement imperative
                          #:location (.location on))))
      (map make-on triggers)))
  (if (null? (ast:statement* o)) '()
      (let ((imperatives (tree:collect
                          o
                          ast:imperative?
                          #:stop?
                          (conjoin (is? <ast>)
                                   (disjoin ast:declarative?
                                            (compose ast:declarative?
                                                     tree:parent))))))
        (append-map imperative->canonical-ons imperatives))))

(define-method (canonical-on->guard (o <canonical-on>))
  (let* ((guard (.guard o))
         (trigger (.trigger o))
         (triggers (make <triggers> #:elements (list trigger)))
         (blocking (.blocking o))
         (statement (.statement o))
         (statement (if (not blocking) statement
                        (clone blocking #:statement statement)))
         (statement (make <on> #:triggers triggers #:statement statement
                          #:location (.location o))))
    (clone guard #:statement statement)))

(define (sort-canonical-ons ons)
  (match ons
    (() '())
    ((on rest ...)
     (let* ((trigger (.trigger on))
            (shared rest (partition
                          (compose (cute ast:equal? <> trigger) .trigger)
                          ons)))
       (cons shared (sort-canonical-ons rest))))))

(define* (canonical-ons->on ons #:key otherwise?)
  (define (canonical-on->guard on)
    (let* ((statement (.statement on))
           (blocking (.blocking on))
           (statement (if blocking (clone blocking #:statement statement)
                          statement))
           (guard (.guard on)))
      (clone guard #:statement statement)))
  (let* ((on (car ons))
         (guards (map canonical-on->guard ons))
         (otherwise (make <guard>
                      #:expression (make <otherwise>)
                      #:statement (make <illegal>)))
         (true-guard? (find (compose ast:literal-true? .expression) guards))
         (otherwise? (and otherwise? (not true-guard?)))
         (guards (if (not otherwise?) guards
                     `(,@guards ,otherwise)))
         (trigger (.trigger on))
         (triggers (make <triggers> #:elements (list trigger)))
         (location (.location on))
         (on (make <on> #:triggers triggers #:location location))
         (statement (make <compound> #:elements guards #:location location)))
    (clone on #:statement statement)))


;;;
;;; Canonical-on normalizations.
;;;
(define (make-declarative-illegal/illegal trigger)
  (make (cond ((or (tree:ancestor trigger <interface>)
                   (ast:provides? trigger))
               <declarative-illegal>)
              (else
               <illegal>))))

(define* (implicit-illegal->explicit-illegal
          trigger ons #:key (make-illegal (const (make <illegal>))))
  (let* ((ons (filter (compose (cute ast:equal? <> trigger) .trigger) ons))
         (guard (and-not-guards (map .guard ons)))
         (guard (simplify-guard guard)))
    (if (ast:literal-false? (.expression guard)) ons
        (let* ((provides? (and=> (.port trigger) ast:provides?))
               (model (tree:ancestor trigger <model>))
               (location (or (and (pair? ons) (car ons))
                             model))
               (on (graft model (make <canonical-on>
                                  #:guard guard
                                  #:trigger trigger
                                  #:statement (make-illegal trigger)
                                  #:location (.location location)))))
          (append ons (list on))))))

(define-method (model->triggers (o <interface>))
  (define (event->trigger event)
    (graft event (make <trigger> #:event.name (.name event)
                       #:location (.location o))))
  (map event->trigger (ast:in-event* o)))

(define-method (model->triggers (o <component-model>))
  (define (port+event->trigger port event)
    (graft o (make <trigger>
               #:port.name (.name port)
               #:event.name (.name event)
               #:location (.location o))))
  (define (port->triggers o)
    (map (cute port+event->trigger o <>)
         ((if (ast:provides? o) ast:in-event*
              ast:out-event*)
          (.type o))))
  (append-map port->triggers (ast:port* o)))

(define* (implicit-illegals->explicit-illegals
          model ons #:key (make-illegal (const (make <illegal>))))
  (let ((modeling (filter (compose ast:modeling? .trigger) ons))
        (triggers (model->triggers model)))
    (append
     (append-map (cut implicit-illegal->explicit-illegal <> ons
                      #:make-illegal make-illegal)
                 triggers)
     modeling)))

(define-method (add-the-end (o <canonical-on>))
  (let ((statement (.statement o)))
    (if (ast:illegal? statement) o
        (let ((statement (ast:add-statement statement (make <the-end>))))
          (clone o #:statement statement)))))

(define-method (add-void-reply (o <canonical-on>))
  (let* ((trigger (.trigger o))
         (port (.port trigger))
         (statement (.statement o))
         (provides? (and port (ast:provides? port)))
         (model (tree:ancestor o <model>))
         (reply? (and (not (ast:typed? trigger))
                      (not (.blocking o))
                      (or (and (is-a? model <interface>)
                               (not (ast:modeling? trigger)))
                          (and provides?
                               (null? (tree:collect
                                       statement
                                       (conjoin
                                        (is? <reply>)
                                        (disjoin (negate .port)
                                                 (compose (cute eq? port <>)
                                                          .port))))))))))
    (if (not reply?) o
        (let ((statement (ast:add-statement* (.statement o) (make <reply>))))
          (clone o #:statement statement)))))

(define-method (illegal->declarative-illegal (o <canonical-on>))
  (let ((declarative? (and (ast:illegal? (.statement o))
                           (or (is-a? (tree:ancestor o <model>) <interface>)
                               (ast:provides? (.trigger o))))))
    (if (not declarative?) o
        (graft o #:statement (make <declarative-illegal>)))))

(define-method (with-behavior (o <behavior>) procedure)
  (lambda (x)
    (let ((x (procedure x)))
      (match x
        ((ons ...) (map (cute graft* o <>) ons))
        (o (graft* o x))))))

(define (normalize:state o)
  "Push guards up, thereby splitting the body of a trigger into multiple
guarded occurrences."
  (match o
    (($ <behavior>)
     (clone o
            #:statement
            ((compose
              (cute make <declarative-compound> #:elements <>)
              (cute map canonical-on->guard <>)
              (cute map (cute group-expressions <>
                              (list <and> <field-test> <or>))
                    <>)
              (with-behavior o (cute map simplify-guard <>))
              statement->canonical-on
              .statement
              ) o)
            #:functions
            (group-expressions (.functions o) (list <and> <field-test> <or>))))
    ((? (%normalize:short-circuit?))
     o)
    ((? (is? <ast>))
     (tree:shallow-map normalize:state o))
    (_
     o)))

(define-method (make-a=>b-expression (a <expression>) (b <expression>))
  (make <or> #:left (make <not> #:expression a) #:right b))

(define-method (invariant:include-guard-expressions (o <invariant>))
  (let ((guards (filter (is? <guard>) (tree:path o))))
    (if (null? guards) o
        (let* ((expressions (map .expression guards))
               (parent (tree:parent o))
               (expression (make-a=>b-expression (and-expressions expressions)
                                                 (.expression o))))
          (graft o #:expression expression)))))

(define %literal-true (make <literal> #:value "true"))

(define-method (make-behavior-invariant (o <behavior>))
  (let* ((invariants (tree:collect o (is? <invariant>)))
         (invariants (map invariant:include-guard-expressions invariants))
         (invariants (map (cute group-expressions <> (list <and> <field-test> <or>)) invariants))
         (expression (and-expressions (map .expression invariants))))
     (make <invariant> #:expression expression)))

(define (normalize:state+illegals o)
  (match o
    (($ <behavior>)
     (let ((model (tree:ancestor o <model>)))
       (clone o
              #:statement
              ((compose
                (cute make <declarative-compound> #:elements <>)
                (cute map canonical-on->guard <>)
                (cute map (cute group-expressions <>
                                (list <and> <field-test> <or>))
                      <>)
                (with-behavior o (cute map simplify-guard <>))
                (cut implicit-illegals->explicit-illegals model <>
                     #:make-illegal make-declarative-illegal/illegal)
                (cute map add-the-end <>)
                (cute map add-void-reply <>)
                (cute map illegal->declarative-illegal <>)
                (cut statement->canonical-on <>)
                (cute tree:transform <> (conjoin (is? <guard>)
                                                 (negate .statement))
                      (const #f))
                (cute tree:transform <> (is? <invariant>) (const #f))
                .statement)
               o)
              #:functions
              (group-expressions (.functions o) (list <and> <field-test> <or>))
              #:invariant (make-behavior-invariant o))))
    ((? (%normalize:short-circuit?))
     o)
    ((? (is? <ast>))
     (tree:shallow-map normalize:state+illegals o))
    (_
     o)))

(define (normalize:event o)
  "Merge all occurrences of a trigger into a single unguarded `on',
i.e., pushing guards into the body of the trigger."
  (match o
    (($ <behavior>)
     (clone o
            #:statement
            ((compose
              (cute make <compound> #:elements <>)
              (cute map canonical-ons->on <>)
              sort-canonical-ons
              (cute map (cute group-expressions <>
                              (list <and> <field-test> <or>))
                    <>)
              (with-behavior o (cute map simplify-guard <>))
              (cute map alpha-rename <>)
              (with-behavior o (cute map formal-dereference <>))
              statement->canonical-on
              .statement
              ) o)
            #:functions
            (group-expressions (.functions o) (list <and> <field-test> <or>))))
    ((? (%normalize:short-circuit?))
     o)
    ((? (is? <ast>))
     (tree:shallow-map normalize:event o))
    (_
     o)))

(define (normalize:event+illegals o)
  (match o
    (($ <behavior>)
     (let ((model (tree:ancestor o <model>)))
       (clone o
              #:statement
              ((compose
                (cute make <compound> #:elements <>)
                (cute map (cut canonical-ons->on <> #:otherwise? #t) <>)
                sort-canonical-ons
                (cute map (cute group-expressions <>
                                (list <and> <field-test> <or>))
                      <>)
                (with-behavior o (cute map simplify-guard <>))
                (cute map alpha-rename <>)
                (cute implicit-illegals->explicit-illegals model <>)
                (with-behavior o (cute map formal-dereference <>))
                statement->canonical-on
                .statement
                ) o)
              #:functions
              (group-expressions (.functions o) (list <and> <field-test> <or>)))))
    ((? (%normalize:short-circuit?))
     o)
    ((? (is? <ast>))
     (tree:shallow-map normalize:event+illegals o))
    (_
     o)))

(define-method (alpha-rename (o <canonical-on>))
  "Rewrite names of formals and locals for event-first normalization.

Event-first normalization groups all guarded event handling blocks for a
specific event in one @code{on EVENT @{@dots{}@}} clause.  Before
normalization each block can use its own naming for the parameters as
formally declared in the interface.  Normalization implies that these
specific parameter names have to be unified.

We follow the following renaming strategy:

@itemize
@item
formals are rewritten to their name as declared in the interface

@item
when the formal name clashes with a member variable, append as many
'x's to the formal name to avoid clashes

@item
if the new formal name clashes with a local, append an 'x' to the local
to prevent unintended shadowing
@end itemize"

  (define (pair-equal? p) (equal? (car p) (cdr p)))

  (define ((rename mapping) o)
    (define (formal-or-local? o)
      (or (is-a? o <formal>)
          (and (is-a? o <variable>)
               (not (is-a? (tree:parent o) <variables>)))))
    (define (rename-string o)
      (or (assoc-ref mapping o) o))
    (match o
      ((and ($ <trigger>) (? (compose null? ast:formal*)))
       o)
      (($ <trigger>)
       (let ((elements (map (rename mapping) (ast:formal* o))))
         (clone o #:formals (clone (.formals o) #:elements elements))))
      (($ <action>)
       (clone o #:arguments ((rename mapping) (.arguments o))))
      (($ <arguments>)
       (clone o #:elements (map (rename mapping) (ast:argument* o))))
      (($ <reference>)
       (if (not (formal-or-local? (.variable o))) o
           (clone o #:name (rename-string (.name o)))))
      (($ <assign>)
       (let ((expression ((rename mapping) (.expression o)))
             (name (if (not (formal-or-local? (.variable o))) (.variable.name o)
                       (rename-string (.variable.name o)))))
         (clone o #:variable.name name #:expression expression)))
      (($ <variable>)
       (let ((name (if (formal-or-local? o) (rename-string (.name o))
                       (.name o))))
         (clone o #:name name #:expression ((rename mapping) (.expression o)))))
      (($ <field-test>)
       (if (not (formal-or-local? (.variable o))) o
           (clone o #:name (rename-string (.variable.name o)))))
      (($ <formal>)
       (clone o #:name (rename-string (.name o))))
      (($ <formal-binding>)
       (clone o #:name (rename-string (.name o))))
      (($ <interface>)
       o)
      ((? (%normalize:short-circuit?))
       o)
      ((? (is? <ast>))
       (tree:shallow-map (rename mapping) o))
      (_
       o)))

  (let* ((trigger (.trigger o))
         (event (.event trigger))
         (formals ((compose ast:formal* .signature) event))
         (formals-ok? (or (pair? (ast:formal* trigger))
                          (null? formals)))
         (model (tree:ancestor o <model>))
         (o (if formals-ok? o
                (let* ((formals (clone (.formals trigger) #:elements formals))
                       (trigger (clone trigger #:formals formals)))
                  (clone o #:trigger trigger))))
         (trigger (.trigger o))
         (trigger (if (pair? (ast:formal* trigger)) trigger
                      (let* ((formals (ast:formal* event))
                             (formals (clone (.formals trigger)
                                             #:elements formals)))
                        (clone trigger #:formals formals))))
         (formals (map .name ((compose .elements .formals .signature) event)))
         (members (map .name (ast:variable* model)))
         (locals (map .name (tree:collect (.statement o) (is? <variable>))))
         (occupied members)
         (fresh (letrec ((fresh (lambda (occupied name)
                                  (if (member name occupied)
                                      (fresh occupied (string-append name "x"))
                                      name))))
                  fresh)) ;; occupied name -> namex
         (refresh (lambda (occupied names)
                    (fold-right (lambda (name o)
                                  (cons (fresh o name) o))
                                ;; occupied names -> (append namesx occupied)
                                occupied names)))
         (fresh-formals (list-head (refresh occupied formals) (length formals)))
         (mapping (filter
                   (negate pair-equal?)
                   (map cons
                        (map .name ((compose .elements .formals) trigger))
                        fresh-formals)))
         (occupied (append (map cdr mapping) members))
         (mapping (append (map cons
                               locals
                               (list-head (refresh occupied locals)
                                          (length locals)))
                          mapping)))
    (if (null? mapping) o
        (clone o
               #:trigger ((rename mapping) trigger)
               #:statement ((rename mapping) (.statement o))))))


;;;
;;; Root normalizations.
;;;
(define-method (purge-data (o <root>))
  "Remove every `extern' data variable and reference."
  (define (parent-location o)
    (or (.location o)
        (parent-location (tree:parent o))))
  (define (keep-call-or-action o)
    (and (or (is-a? o <assign>) (is-a? o <variable>) (is-a? o <reply>))
         (or (is-a? (.expression o) <call>) (is-a? (.expression o) <action>))
         (.expression o)))
  (define (extern-signature o)
    (and (is-a? o <signature>) (is-a? (.type o) <extern>)))
  (define (make-void-signature o)
    (and (is-a? o <signature>) (clone o #:type.name "void")))
  (define (extern-return o)
    (and (is-a? o <return>)
         (or (not (.expression o))
             (is-a? (ast:type o) <extern>))))
  (define (make-void-return o)
    (and (is-a? o <return>) (make <return> #:location (.location o))))
  (define (extern-reply o)
    (and (is-a? o <reply>)
         (or (not (.expression o))
             (is-a? (ast:type o) <extern>))))
  (define (make-void-reply o)
    (and (is-a? o <reply>) (make <reply> #:location (.location o))))
  (let* ((data? (disjoin (is? <extern>)
                         (is? <data-expr>)
                         (conjoin (disjoin (is? <assign>)
                                           (is? <formal>)
                                           (is? <formal-reference>)
                                           (is? <reference>)
                                           (is? <variable>))
                                  (compose (is? <extern>) ast:type))))
         (data->skip-or-#f (conjoin (is? <statement>)
                                    (compose not (is? <variables>) tree:parent)
                                    (compose not (is? <compound>) tree:parent)
                                    (compose (cute make <skip> #:location <>)
                                             parent-location
                                             tree:parent)))
         (transform (disjoin keep-call-or-action
                             make-void-signature
                             make-void-return
                             make-void-reply
                             data->skip-or-#f)))
    (tree:transform o (disjoin data?
                               extern-signature
                               extern-return
                               extern-reply)
                    transform)))

(define-method (add-void-function-return (o <root>))
  "Make implicit returns explicit for void functions."
  (define* (add-return o #:key (loc o))
    (match o
      (($ <function>)
       (clone o #:statement (add-return (.statement o) #:loc o)))
      (($ <compound>)
       (clone o #:elements (add-return (ast:statement* o) #:loc o)))
      ((statement ... t)
       (append o (list (make <return> #:location (.location (tree:parent t))))))
      ((statement ...)
       (append o (list (make <return> #:location (.location loc)))))))
  (define void-function-with-implicit-return?
    (conjoin (is? <function>)
             (compose (is? <void>) ast:type)))
  (tree:transform o void-function-with-implicit-return? add-return))


(define-method (extract-call (o <root>))
  "Move typed function calls from variable initialization and assignment
to a separate statement, for mCRL2."
  (define (extract-assign/variable-call o)
    (let ((expression (make <literal> #:value "return_value"))
          (call (.expression o)))
      (match o
        (($ <assign>)
         (let ((assign (clone o #:expression expression)))
           (list call assign)))
        (($ <variable>)
         (let* ((default (ast:default-value o))
                (variable (clone o #:expression default))
                (parent (tree:parent o))
                (assign (graft parent (make <assign>
                                        #:variable.name (.name variable)
                                        #:expression expression))))
           (list variable call assign))))))

  (define (extract-call+compound o)
    (let* ((statements (extract-assign/variable-call o))
           (compound (make <compound> #:elements statements)))
      (graft (tree:parent o) compound)))

  (define (compound-extract-call o)
    (let ((elements (append-map
                     (lambda (o)
                       (if (not (assign/variable-call? o)) (list o)
                           (extract-assign/variable-call o)))
                     (.elements o))))
      (graft o #:elements elements)))

  (define assign/variable-call?
    (conjoin
     (disjoin (is? <assign>)
              (is? <variable>))
     (compose (is? <call>) .expression)))

  (define compound-call?
    (conjoin (is? <compound>)
             (compose (cute any assign/variable-call? <>)
                      .elements)))

  (define single-call?
    (conjoin assign/variable-call?
             (compose not (is? <compound>) tree:parent)))

  (define predicate-call?
    (conjoin (is? <call>) ast:declarative?))

  (tree:transform
   o
   `((,single-call? . ,extract-call+compound)
     (,compound-call? . ,compound-extract-call))))

(define-method (inline-predicates (o <ast>))
  "Inline declarative (predicates) calls."
  (define predicate-call?
    (conjoin (is? <call>) ast:declarative?))
  (define (inline-predicate-call o)
    (let ((expression
           ((compose .expression last ast:statement* .statement .function) o)))
      (tree:transform expression (is? <call>) inline-predicate-call)))
  (tree:transform o predicate-call? inline-predicate-call))

(define-method (add-defer-end (o <root>))
  (define (add-end o)
    (let* ((end (graft o (make <defer-end>)))
           (statement (ast:add-statement* (.statement o) end)))
      (graft o #:statement statement)))
  (tree:transform o (is? <defer>) add-end))

(define (formal-dereference o)
  (let ((model (tree:ancestor o <model>)))
    (define (rescope o orig interface-formal)
      (let* ((o (graft* (tree:parent interface-formal) o))
             (o (ast:rescope o model)))
        (graft* (tree:parent orig) o)))
    (define (replace-formal-reference o)
      (let* ((interface-formal (.formal o))
             (formal (clone interface-formal #:name (.name o))))
        (rescope formal o interface-formal)))
    (define (replace-formal-reference-binding o)
      (let* ((interface-formal (.formal o))
             (formal (make <formal-binding>
                       #:name (.name o)
                       #:direction (.direction interface-formal)
                       #:type.name (.type.name interface-formal)
                       #:variable.name (.variable.name o))))
        (rescope formal o interface-formal)))
    (tree:transform
     o
     `((,(is? <formal-reference-binding>) . ,replace-formal-reference-binding)
       (,(is? <formal-reference>) . ,replace-formal-reference)))))

(define* (add-reply-port o #:optional (port #f) (block? #f))
  (match o
    (($ <reply>)
     (let ((port? (.port o)))
       (if (and port? (not (string? port?))) o
           (clone o #:port.name (.name port)))))
    (($ <blocking>)
     (if (not block?) (add-reply-port (.statement o) port block?)
         (let ((elements
                (let ((s (.statement o)))
                  (if (not (is-a? s <compound>))
                      (list (add-reply-port s port block?))
                      (map (cute add-reply-port <> port block?)
                           (ast:statement* s))))))
           (make <blocking-compound> #:port port #:elements elements))))
    (($ <on>)
     (let ((statement (add-reply-port
                       (.statement o)
                       (if port port ((compose .port car ast:trigger*) o))
                       ((compose ast:provides? .port car ast:trigger*) o))))
       (clone o #:statement statement)))
    (($ <guard>)
     (clone o #:statement (add-reply-port (.statement o) port block?)))
    (($ <compound>)
     (let ((elements (map (cute add-reply-port <> port block?)
                          (ast:statement* o))))
       (clone o #:elements elements)))
    (($ <behavior>)
     (clone o #:statement (add-reply-port (.statement o) port block?)
            #:functions (add-reply-port (.functions o) port block?)))
    ((? (%normalize:short-circuit?)) o)
    (($ <component>)
     (let ((behavior (add-reply-port (.behavior o)
                                     (and (= 1 (length (ast:provides-port* o)))
                                          (ast:provides-port o))
                                     block?)))
       (clone o #:behavior behavior)))
    (($ <interface>)
     o)
    ((? (is? <ast>))
     (tree:shallow-map (cut add-reply-port <> port block?) o))
    (_
     o)))

(define* ((binding-into-blocking #:optional (locals '())) o)

  (define (formal-binding->formal o)
    (match o
      (($ <formal-binding>)
       (make <formal> #:name (.name o) #:type.name (.type.name o)
             #:direction (.direction o)))
      (_
       o)))

  (define ((passdown-formal-bindings formal-bindings) o)
    (match o
      ((and ($ <compound>) (? ast:declarative?))
       (let ((elements (map (passdown-formal-bindings formal-bindings)
                            (ast:statement* o))))
         (clone o #:elements elements)))
      (($ <declarative-illegal>) o)
      ((? ast:declarative?)
       (let ((statement ((passdown-formal-bindings formal-bindings) (.statement o))))
         (clone o #:statement statement)))
      (($ <compound>)
       (clone o #:elements (cons formal-bindings (ast:statement* o))))
      (_
       (make <compound> #:elements (cons formal-bindings (list o))))))

  (match o
    (($ <on>)
     (let* ((trigger ((compose car ast:trigger*) o))
            (on-formals (ast:formal* trigger))
            (formal-bindings (filter (is? <formal-binding>) on-formals))
            (formal-bindings (and (pair? formal-bindings)
                                  (make <out-bindings>
                                    #:elements formal-bindings
                                    #:port (.port trigger))))
            (on-formals (map formal-binding->formal on-formals)))
       (if (not formal-bindings) o
           (let* ((formals (make <formals> #:elements on-formals))
                  (trigger (clone trigger #:formals formals))
                  (triggers (clone (.triggers o) #:elements (list trigger)))
                  (statement ((passdown-formal-bindings formal-bindings)
                              (.statement o))))
             (clone o #:triggers triggers #:statement statement)))))
    (($ <behavior>)
     (clone o #:statement ((binding-into-blocking '()) (.statement o))))
    ((? (%normalize:short-circuit?))
     o)
    (($ <component>)
     (clone o #:behavior ((binding-into-blocking) (.behavior o))))
    (($ <interface>)
     o)
    ((? (is? <ast>))
     (tree:shallow-map (binding-into-blocking locals) o))
    (_ o
       )))

(define* (remove-otherwise o #:optional (keep-annotated? #t) (statements '()))
  "Replace otherwise with the negated conjunction of every other guard at
the same level."
  (define (virgin-otherwise? x)
    (or (equal? x "otherwise") (eq? x *unspecified*)))
  (match o
    ((? ast:imperative?)
     o)
    ((and ($ <guard>)
          (= .expression (and ($ <otherwise>) (= .value value)))
          (= .statement statement)) (=> failure)
          (if (or (and keep-annotated?
                       (not (virgin-otherwise? value)))
                  (null? statements))
              (failure)
              (clone o
                     #:expression (not-or-guards statements)
                     #:statement (remove-otherwise statement keep-annotated?))))
    ((and ($ <compound>) (= ast:statement* (statements ...)))
     (let ((elements (map (cute remove-otherwise <> keep-annotated? statements)
                          statements) ))
       (clone o #:elements elements)))
    (($ <skip>)
     o)
    (($ <functions>)
     o)
    ((? (%normalize:short-circuit?))
     o)
    ((and (? (is? <component>) (= .behavior behavior)))
     (let ((behavior (remove-otherwise behavior keep-annotated? statements)))
       (clone o #:behavior behavior)))
    ((and (? (is? <interface>) (= .behavior behavior)))
     (let ((behavior (remove-otherwise behavior keep-annotated? statements)))
       (clone o #:behavior behavior)))
    ((? (is? <ast>))
     (tree:shallow-map (cute remove-otherwise <> keep-annotated? statements) o))
    (_
     o)))

(define (remove-location o)
  "Remove locations from types, events, formals, signatures, ports."
  (match o
    (($ <interface>)
     (clone o #:events (remove-location (.events o))))
    (($ <component>)
     (clone o #:ports (remove-location (.ports o))))
    (($ <foreign>)
     (clone o #:ports (remove-location (.ports o))))
    (($ <instance>)
     o)
    (($ <system>)
     (clone o
            #:ports (remove-location (.ports o))
            #:bindings (remove-location (.bindings o))))
    (($ <event>)
     (clone o
            #:location #f
            #:signature (remove-location (.signature o))))
    (($ <port>)
     (clone o
            #:formals (remove-location (.formals o))))
    (($ <binding>)
     (clone o
            #:left (remove-location (.left o))
            #:right (remove-location (.right o))))
    (($ <signature>)
     (clone o
            #:location #f
            #:formals (remove-location (.formals o))))
    (($ <root>)
     (tree:shallow-map remove-location o))
    ((? (is? <namespace>))
     (tree:shallow-map remove-location o))
    ((? (is? <tree:locationed>))
     (clone o #:location #f))
    ((? (is? <ast>))
     (tree:shallow-map remove-location o))
    (_
     o)))

(define (remove-behavior o)
  "Remove behavior from models."
  (match o
    (($ <interface>)
     (clone o #:behavior #f))
    (($ <component>)
     (clone o #:behavior #f))
    ((? (is? <namespace>))
     (tree:shallow-map remove-behavior o))
    (_
     o)))

(define (replace-expression old o var)
  (cond ((eq? old o)
         var)
        ((is-a? o <expression>)
         (tree:shallow-map (cute replace-expression old <> var) o))
        ((is-a? o <arguments>)
         (tree:shallow-map (cute replace-expression old <> var) o))
        (else
         o)))

;; FIXME: add-explicit-temporaries
;; this function is overly complicated:
;; - it has too many nested helper functions
;; - which are too dependend
;; - there is too much duplication
;; - and too much (mutual) recursion

(define ((split-complex? temporaries add-temporary?) o)
  (and (pair? (tree:collect o (disjoin (is? <and>) (is? <or>))))
       (pair? (temporaries o))
       (not (add-temporary? o))))

(define ((split-complex-expressions split-complex?) o)
  "Split && and || into an if with a simple expression.  Depends on the
add-explicit-temporaries transformation for splitting argument lists."
  (let ((split-complex-expressions (split-complex-expressions split-complex?)))
    (define* (split o #:key not-e)
      (let* ((parent (tree:parent o))
             (expression (.expression o))
             (expression (simplify-expression expression))
             (o (graft* parent o #:expression expression)))
        (match expression
          ((and ($ <and>) (= .left left) (= .right right))
           (let* ((then (cond (not-e
                               (.then o))
                              (else
                               (clone o #:expression right))))
                  (false (make <literal> #:value "false"))
                  (left (if (not not-e) left
                            (clone not-e #:expression left)))
                  (left (simplify-expression left))
                  (else (cond ((and not-e (is-a? o <if>))
                               (clone o #:expression
                                      (clone not-e #:expression right)))
                              ((is-a? o <if>)
                               (.else then))
                              (else
                               (clone o #:expression false))))
                  (simple (graft parent (make <if>
                                          #:expression left
                                          #:then then
                                          #:else else
                                          #:location (.location o)))))
             (split-complex-expressions simple)))
          ((and (? (const not-e))
                ($ <or>) (= .left left) (= .right right))
           (let* ((then (cond (not-e
                               (clone o #:expression
                                      (clone not-e #:expression right)))
                              (else
                               (clone o #:expression right))))
                  (false (make <literal> #:value "false"))
                  (left (if (not not-e) left
                            (clone not-e #:expression left)))
                  (left (simplify-expression left))
                  (else (cond ((is-a? o <if>)
                               (.else then))
                              (else
                               (clone o #:expression false))))
                  (simple (graft parent (make <if>
                                          #:expression left
                                          #:then then
                                          #:else else
                                          #:location (.location o)))))
             (split-complex-expressions simple)))
          ((and ($ <or>) (= .left left) (= .right right))
           (let* ((true (make <literal> #:value "true"))
                  (then (cond ((is-a? o <if>)
                               (.then o))
                              (else
                               (clone o #:expression true))))
                  (else (clone o #:expression right))
                  (left (simplify-expression left))
                  (simple (graft parent (make <if>
                                          #:expression left
                                          #:then then
                                          #:else else
                                          #:location (.location o)))))
             (split-complex-expressions simple)))
          (($ <group>)
           (let* ((expression-expression (.expression expression))
                  (o (graft o #:expression expression-expression)))
             (split o #:not-e not-e)))
          (($ <not>)
           (let* ((expression-expression (.expression expression))
                  (o (graft o #:expression expression-expression)))
             (split o #:not-e expression)))
          ((? (const not-e))
           (let* ((expression (.expression o))
                  (group? (or (is-a? expression <binary>)
                              (is-a? expression <field-test>)))
                  (expression (if (not group?) expression
                                  (make <group> #:expression expression)))
                  (expression (clone not-e #:expression expression)))
             (graft o #:expression expression)))
          (_
           o))))

    (define split?
      (conjoin (disjoin (is? <if>)
                        (is? <assign>)
                        (is? <reply>)
                        (is? <return>)
                        (is? <variable>))
               split-complex?))

    (define (split+ o)
      (match o
        (($ <variable>)
         (let* ((parent (tree:parent o))
                (expression (.expression o))
                (expression (simplify-expression expression))
                (o (graft* parent o #:expression expression))
                (variable assign (split-variable o))
                (o (split (graft assign #:expression expression)))
                (parent (tree:ancestor o <statement>)))
           (graft parent (make <compound>
                           #:elements (list variable o)
                           #:location (.location o)))))
        (_ (split o))))

    (tree:transform o split? split+)))

(define* (normalize:split-complex-expressions #:key call-only?)
  (if call-only? (split-complex-expressions
                  (split-complex? temporaries* add-temporary*?))
      (split-complex-expressions
       (split-complex? temporaries add-temporary?))))

(define ((add-temporary add-temporary?) o)
  (let* ((expression (.expression o))
         (variable-expression (add-temporary? o))
         (type (ast:type variable-expression))
         (void? (is-a? type <void>))
         (local-type? (eq? (tree:ancestor type <model>)
                           (tree:ancestor o <model>)))
         (type-name (cond
                     ((is-a? type <subint>) (.name (ast:type (make <int>))))
                     (local-type? (.name type))
                     (else (ast:dotted-name type))))
         (name (temp-name o))
         (location (.location o))
         (behavior (tree:ancestor o <behavior>))
         (parent (or (tree:ancestor o <statement>)
                     (tree:ancestor o <behavior>)))
         (temporary (make <variable>
                      #:name name
                      #:type.name type-name
                      #:expression variable-expression
                      #:location location))
         (temporary (if void? variable-expression temporary))
         (var (make <reference> #:name name #:location location))
         (var (if void? #f var))
         (o (tree:shallow-map
             (cute replace-expression variable-expression <> var)
             o)))
    (graft parent (make <compound>
                    #:elements (list temporary o)
                    #:location (.location o)))))

(define ((split+add-temporaries split-complex? split-complex-expressions
                                add-temporary add-explicit-temporaries) o)
  (let ((split+add-temporaries (split+add-temporaries
                                split-complex? split-complex-expressions
                                add-temporary add-explicit-temporaries)))
    (if (not (split-complex? o)) (add-temporary o)
        (let ((o (split-complex-expressions o)))
          (if (add-temporary? o) (add-explicit-temporaries o)
              o)))))

(define* ((add-explicit-temporaries #:key call-only?) o)
  "Make implicit temporary values in action, call, if, reply, and return
expressions explicit."
  (let* ((add-explicit-temporaries (add-explicit-temporaries
                                    #:call-only? call-only?))
         (add-temporary? (if call-only? add-temporary*? add-temporary?))
         (temporaries (if call-only? temporaries* temporaries))
         (split-complex? (split-complex? temporaries add-temporary?))
         (split-complex-expressions (split-complex-expressions split-complex?))
         (add-temporary (add-temporary add-temporary?))
         (split+add-temporaries (split+add-temporaries
                                 split-complex? split-complex-expressions
                                 add-temporary add-explicit-temporaries)))
    (match o
      (($ <if>)
       (let* ((expression (.expression o))
              (split-expression? (or (add-temporary? (.expression o))
                                     (split-complex? (.expression o)))))
         (if split-expression? (let ((o (split+add-temporaries o)))
                                 (add-explicit-temporaries o))
             (let* ((then (add-explicit-temporaries (.then o)))
                    (else (add-explicit-temporaries (.else o)))
                    (location (.location o))
                    (o (graft o
                              #:expression expression
                              #:then then
                              #:else else)))
               (if (not (add-temporary? o)) o
                   (split+add-temporaries o))))))
      ((or ($ <call>) ($ <action>))
       (let ((split? (or (split-complex? o) (add-temporary? o))))
         (if (not split?) o
             (let ((o (split+add-temporaries o)))
               (add-explicit-temporaries o)))))
      ((or ($ <assign>) ($ <reply>) ($ <return>) ($ <variable>))
       (let ((split? (or (split-complex? o) (add-temporary? o))))
         (if (not split?) o
             (let ((o (split+add-temporaries o)))
               (add-explicit-temporaries o)))))
      (($ <defer>)
       (let* ((statement (.statement o))
              (statement (add-explicit-temporaries statement)))
         (clone o #:statement statement)))
      ((and ($ <compound>) (? ast:declarative?))
       (clone o #:elements (map add-explicit-temporaries (ast:statement* o))))
      (($ <compound>)
       (let ((statements
              (let loop ((statements (ast:statement* o)))
                (match statements
                  (()
                   '())
                  ((statement rest ...)
                   (match statement
                     ((or ($ <compound>) ($ <defer>) ($ <if>))
                      (cons (add-explicit-temporaries statement) (loop rest)))
                     ((or (? add-temporary?) (? split-complex?))
                      (let ((split (add-explicit-temporaries statement)))
                        (match split
                          (($ <compound>)
                           (append (.elements split) (loop rest)))
                          ((? (is? <statement>))
                           (cons split (loop rest))))))
                     (_
                      (cons statement (loop rest)))))))))
         (clone o #:elements statements)))
      (($ <function>)
       (clone o #:statement (add-explicit-temporaries (.statement o))))
      (($ <behavior>)
       (clone o
              #:functions (add-explicit-temporaries (.functions o))
              #:statement (add-explicit-temporaries (.statement o))))
      ((? (%normalize:short-circuit?))
       o)
      (($ <interface>)
       (clone o #:behavior (add-explicit-temporaries (.behavior o))))
      (($ <component>)
       (clone o #:behavior (add-explicit-temporaries (.behavior o))))
      ((? (is? <ast>)) (tree:shallow-map add-explicit-temporaries o))
      (_ o))))

(define* (group-expressions o #:optional (group (list)))
  (match o
    ((? (const (find (cute is-a? o <>) group)))
     (let* ((parent (tree:parent o))
            (o (tree:shallow-map (cute group-expressions <> group) o))
            (o (graft* parent o)))
       (match (tree:parent o)
         (($ <group>) o)
         ((? (is? <expression>)) (make <group> #:expression o))
         (else o))))
    ((and (? (is? <not>))
          (= .expression (and expression (or (? (is? <binary>))
                                             (? (is? <field-test>))))))
     (let ((expression (tree:shallow-map (cute group-expressions <> group) expression)))
       (graft o #:expression (make <group> #:expression expression))))
    (($ <canonical-on>)
     (let* ((guard (group-expressions (.guard o) group))
            (statement (group-expressions (.statement o) group)))
       (graft o #:guard guard #:statement statement)))
    (($ <function>)
     (clone o #:statement (group-expressions (.statement o) group)))
    (($ <behavior>)
     (clone o
            #:functions (group-expressions (.functions o) group)
            #:statement (group-expressions (.statement o) group)))
    ((? (%normalize:short-circuit?))
     o)
    (($ <interface>)
     (clone o #:behavior (group-expressions (.behavior o) group)))
    (($ <component>)
     (clone o #:behavior (group-expressions (.behavior o) group)))
    ((? (is? <ast>))
     (tree:shallow-map (cute group-expressions <> group) o))
    (_
     o)))

(define (simplify-guard-expressions o)
  "Simplify guard expressions by using static analysis."
  (match o
    (($ <behavior>)
     (clone o
            #:statement
            ((compose
              (cute make <declarative-compound> #:elements <>)
              (cute map canonical-on->guard <>)
              (cute map (cute group-expressions <>
                              (list <and> <field-test> <or>))
                    <>)
              (cute map simplify-guard <>)
              statement->canonical-on
              .statement
              ) o)))
    ((? (%normalize:short-circuit?))
     o)
    ((? (is? <ast>))
     (tree:shallow-map simplify-guard-expressions o))
    (_
     o)))

(define (tag-imperative-blocks o)
  "Mark imperative statement blocks with a unique tag for unreachable
code check."
  (define (tag-function? o)
    (not (eq? (.context-called o) 'declarative)))
  (define (location o)
    (or (.location o)
        (location (tree:parent o))))
  (define (add-tag-imperative o)
    (if (ast:imperative? o)
        (match o
          ((? ast:illegal?)
           o)
          ;; FIXME: make $ goops base class aware
          ;; to derive <declarative-compound> from <component>
          ((or ($ <compound>) ($ <declarative-compound>))
           ;; FIXME: (car (ast:trigger* ...
           (let* ((location
                   (location (if (not (is-a? (tree:parent o) <on>)) o
                                 (car (ast:trigger* (tree:parent o))))))
                  (tag (make <tag> #:location location)))
             (graft o #:elements (cons tag (ast:statement* o)))))
          (_
           (let* ((location (location o))
                  (tag (make <tag> #:location location))
                  (elements (list tag o)))
             (graft (tree:parent o)
                    (make <compound>
                      #:elements elements #:location location)))))
        o))

  (match o
    ((or ($ <blocking>) ($ <defer>) ($ <guard>) ($ <on>))
     (let* ((statement (.statement o))
            (statement (add-tag-imperative (tag-imperative-blocks statement))))
       (graft o #:statement statement)))
    (($ <function>)
     (if (not (tag-function? o)) o
         (let* ((statement (.statement o))
                (statement (add-tag-imperative
                            (tag-imperative-blocks statement))))
           (graft o #:statement statement))))
    (($ <if>)
     (let ((then (add-tag-imperative (tag-imperative-blocks (.then o))))
           (else (and=> (tag-imperative-blocks (.else o)) add-tag-imperative)))
       (graft o #:then then #:else else)))
    ((or ($ <compound>) ($ <declarative-compound>))
     (graft o #:elements (map tag-imperative-blocks (ast:statement* o))))
    ((? (is? <statement>))
     o)
    (($ <behavior>)
     (clone o #:statement (tag-imperative-blocks (.statement o))
            #:functions (tag-imperative-blocks (.functions o))))
    (($ <functions>)
     (clone o #:elements (map tag-imperative-blocks (.elements o))))
    ((? (%normalize:short-circuit?))
     o)
    (($ <component>)
     (clone o #:behavior (tag-imperative-blocks (.behavior o))))
    (($ <interface>)
     (clone o #:behavior (tag-imperative-blocks (.behavior o))))
    ((? (is? <type>))
     o)
    ((? (is? <ast>))
     (tree:shallow-map tag-imperative-blocks o))
    (_ o)))

(define %specialized-functions '())
(define %specialized-functions-names '())
(define (add-specialized-function function actuals-port)
  (let* ((formals-port (map .name (filter ast:port? (ast:formal* function))))
         (formal->actual-alist (map cons formals-port actuals-port )))
    (define (rename-name name)
      (or (assoc-ref formal->actual-alist name) name))
    (define (has-port-name? o)
      ((disjoin (is? <action>) (is? <reply>) (is? <shared-reference>) (is? <shared-variable>) (is? <shared-field-test>)) o))
    (define (replace-port-name port)
      (graft port #:port.name (rename-name (.port.name port))))
    (define (port-reference? o)
       (and (is-a? o <reference>) (ast:port? o)))
    (define (replace-reference reference)
      (graft reference #:name (rename-name (.name reference))))
    (let* ((name (string-append (.name function) "''" (string-join actuals-port "'"))))
      (if (find (cute equal? name <>) %specialized-functions-names) name
          (begin
            (set! %specialized-functions-names (cons name %specialized-functions-names))
            (let* ((specialized-function (tree:transform function
                                                         `((,has-port-name? . ,replace-port-name)
                                                         (,port-reference? . ,replace-reference))))
                   (specialized-function (specialize-calls specialized-function))
                   (signature (.signature specialized-function))
                   (formals (graft (.formals signature) #:elements (filter (negate ast:port?) (ast:formal* signature))))
                   (signature (graft signature #:formals formals))
                   (specialized-function (graft specialized-function
                                                #:name name
                                                #:signature signature)))
              (set! %specialized-functions (cons specialized-function %specialized-functions))
              name))))))

(define-method (has-no-port-formals? (o <function>))
  (null? (filter ast:port? (ast:formal* o))))

(define-method (has-port-arguments? (o <call>))
  (and (pair? (filter ast:port? (ast:argument* o)))
       (null? (filter .variable (filter ast:port? (ast:argument* o))))))

(define-method (specialize-call (o <call>))
  (let* ((function (.function o))
         (actuals-port (map .name (filter ast:port? (ast:argument* o))))
         (name (add-specialized-function function actuals-port))
         (arguments (filter (negate ast:port?) (ast:argument* o))))
    (graft o
           #:function.name name
           #:arguments (graft (.arguments o) #:elements arguments))))

(define (specialize-calls o)
  (tree:transform o (conjoin (is? <call>) has-port-arguments?) specialize-call))

(define-method (add-function-port-specializations (o <behavior>))
  (set! %specialized-functions '())
  (set! %specialized-functions-names '())
  (let* ((behavior (specialize-calls o))
         (functions (filter has-no-port-formals? (ast:function* behavior)))
         (functions (graft (.functions behavior) #:elements (append functions %specialized-functions)))
         (behavior (graft behavior #:functions functions)))
    behavior))

(define-method (add-function-port-specializations (o <root>))
  (tree:transform o (is? <behavior>) add-function-port-specializations))

(define-method (global+local-function* (o <behavior>))
  (define (call* o)
    "Return a list of <call>s occurring in ast O"
    (tree:collect o (conjoin
                     (is? <call>)
                     (negate (cute tree:ancestor <> <defer>)))))
  (define (function* c*)
    "Return all unique functions called by C*, a list of calls"
    (delete-duplicates (filter (is? <function>) (map .function c*)) eq?))
  (define (called-function* f)
    "Return the functions of the nested calls in function F"
    (function* (call* (.statement f))))
  (define (global-function? o)
    (not (tree:ancestor o <behavior>)))
  (define (call-closure f*)
    "Return the transitive closure of all the functions of the nested calls
in F*, a list of functions"
    (let loop ((f* f*))
      (let* ((step-f* (append f* (append-map called-function* f*)))
             (step-f* (delete-duplicates step-f* eq?)))
        (if (eq? (length step-f*) (length f*)) f*
            (loop step-f*)))))
  (let* ((calls (call* o))
         (functions (append (filter-map .function calls)))
         (functions (filter global-function? functions))
         (functions (delete-duplicates functions ast:name-equal?))
         (functions (call-closure functions)))
    (lset-union eq? (ast:function* (.functions o)) functions)))

(define (global-functions->behavior ast)
  (tree:transform ast (is? <functions>)
                  (lambda (o)
                    (let* ((functions (global+local-function* (tree:ancestor o <behavior>))))
                      (graft o #:elements functions)))))
